#include <math.h>

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jdrest.h"


/* Implementation of the method developed by: François Alter, Sylvain Durand and
   Jacques Froment:
     Adapted Total Variation for Artifact Free Decompression of JPEG Images
   from Journal of Mathematical Imaging and Vision 23: 199-211, 2005

   Notes:
   - Big part of the implementation has been freely given by Jacques Froment.
   - This method (or implementation) noticeable decreases the image contrast.
*/

#define MAXSAMP 255.
#define MINSAMP 0.

LOCAL(void)
froment_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
min_TV(j_decompress_ptr cinfo, struct jpeg_restoration *rest, int niterations,
    double step, boolean cnt_step, RSAMPBUFFER out);
LOCAL(void)
project(struct jpeg_restoration *rest, RSAMPBUFFER out);

static const float plain_wTV[8]={1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0}; /* Plain TV */
static const float default_wTV[8]={5.0,2.0,1.0,1.0,1.0,2.0,5.0,10.0}; 

struct my_rest_struct
{
  struct jpeg_restoration pub;
  RSAMPBUFFER bitmap;
  RBLOCK qtable;

  RCOEFBUFFER DCTmin, DCTmax, DCTimage;
  int width_in_blocks, height_in_blocks;
  int w, h; /* _blocks * 8 */

  /* Parameters of the algorithm */
  float wTV[8]; /* Weights */
  float step_size;
  int steps;
  boolean constant_step;
};

typedef struct my_rest_struct * my_rest_ptr;

/* Use jpeg_get_large(cinfo, size) */

METHODDEF(void) froment_decompress_plane
  JPP((j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_buf, JBLOCKARRAY blockarray));

GLOBAL(struct jpeg_restoration *)
jpeg_rest_froment(j_decompress_ptr cinfo, J_REST_TARGET target)
{
  my_rest_ptr rest;

  if (target != JREST_TARGET_IDCT_Y &&
    target != JREST_TARGET_IDCT_Cb &&
    target != JREST_TARGET_IDCT_Cr)
  {
    ERREXIT(cinfo, JERR_REST_NOT_APPLICABLE);
  }

  /* The first restoration */
  rest = (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
                                SIZEOF(struct my_rest_struct));

  rest->pub.decompress_plane = froment_decompress_plane;
  rest->pub.target = target;
  add_restorator(cinfo, (struct jpeg_restoration *) rest);
  return (struct jpeg_restoration *) rest;
}

METHODDEF(void)
froment_decompress_plane(j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_ptr, JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,l;
  int buffer_row, block_rows;
  jpeg_component_info * compptr;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION output_col;
  RSAMP value;

  compptr = &cinfo->comp_info[myrest->pub.component];

  if (cinfo->output_iMCU_row == 0)
  {
    froment_restoration(cinfo, rest, blockarray);
  }

  /* Give the rows */

  buffer_row = cinfo->output_iMCU_row * compptr->v_samp_factor;
  
  /* How many rows we should give? */
  if (cinfo->output_iMCU_row < last_iMCU_row)
    block_rows = compptr->v_samp_factor;
  else {
    /* NB: can't use last_row_height here; it is input-side-dependent! */
    block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (block_rows == 0) block_rows = compptr->v_samp_factor;
  }

  for (j=0; j<block_rows; j++)
  {
    output_col = 0;
    for (i=0; i<compptr->width_in_blocks; i++)
    {
      for(k=0; k<8;k++)
        for(l=0; l<8;l++)
        {
          value = myrest->bitmap[((buffer_row + j)*8+k)*myrest->w + (i*8+l)]
	    + 0.5;
          if (value > MAXSAMP)
            (output_ptr[k] + output_col)[l] = MAXSAMP;
          else if (value < MINSAMP)
            (output_ptr[k] + output_col)[l] = MINSAMP;
          else
            (output_ptr[k] + output_col)[l] = value;
            
        }

      output_col += compptr->DCT_scaled_size;
    }
    output_ptr += compptr->DCT_scaled_size;
  }

}

LOCAL(void)
get_env_configuration(j_decompress_ptr cinfo, struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  char *env_var;
  int i;

  myrest->steps = 10;
  myrest->step_size = 1.;
  for (i=0; i<8; i++)
    myrest->wTV[i] = default_wTV[i];

#ifndef NO_GETENV
  env_var = getenv("JPEG_FROMENT_STEPS");
  if (env_var != NULL)
    sscanf(env_var, "%i", &myrest->steps);

  env_var = getenv("JPEG_FROMENT_STEP_SIZE");
  if (env_var != NULL)
    sscanf(env_var, "%g", &myrest->step_size);

  env_var = getenv("JPEG_FROMENT_WEIGHTS");
  if (env_var != NULL)
    sscanf(env_var, "%g%g%g%g%g%g%g%g",
      &myrest->wTV[0],
      &myrest->wTV[1],
      &myrest->wTV[2],
      &myrest->wTV[3],
      &myrest->wTV[4],
      &myrest->wTV[5],
      &myrest->wTV[6],
      &myrest->wTV[7]);

#endif
}

/* !!! Some global functions for setting the variables are needed.
jpeg_rest_froment_set_steps(rest, int)
jpeg_rest_froment_set_step_size(rest, float)
... */

LOCAL(void)
froment_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  jpeg_component_info * compptr;

  compptr = &cinfo->comp_info[myrest->pub.component];

  /* Prepare dimensions */
  myrest->width_in_blocks = compptr->width_in_blocks;
  myrest->height_in_blocks = compptr->height_in_blocks;
  myrest->w = compptr->width_in_blocks * 8;
  myrest->h = compptr->height_in_blocks * 8;

  /* Load the coefficients, and prepare the DCT ranges */
  set_DCT_constraint(cinfo, rest, blockarray);

  /* Allocate the bitmap */
  myrest->bitmap =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Decompress the first image */
  fdct_inverse_image(myrest->DCTimage, myrest->bitmap, myrest->w,
      myrest->h);

  get_env_configuration(cinfo, rest);

  /* Do the minimization */
  min_TV(cinfo, rest, myrest->steps, myrest->step_size, FALSE, myrest->bitmap);
}

LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,row,col, adr;
  static const double aanscalefactor[DCTSIZE] = 
    {
      1.0, 1.387039845, 1.306562965, 1.175875602,
      1.0, 0.785694958, 0.541196100, 0.275899379
    };
  RCOEF *table;

  myrest->DCTmin =
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTmax = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTimage = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);

  if ( (myrest->DCTmin == NULL) || (myrest->DCTmax == NULL) ||
      (myrest->DCTimage == NULL) )
    ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 0);
  
  /* Get quantization table */
  for (i=0; i < DCTSIZE2; i++)
    myrest->qtable[i] = cinfo->comp_info[myrest->pub.component].quant_table->quantval[i];
  table = myrest->qtable;

 
  /* For float AA&N IDCT method, multipliers are equal to quantization
   * coefficients scaled by scalefactor[row]*scalefactor[col], where
   *   scalefactor[0] = 1
   *   scalefactor[k] = cos(k*PI/16) * sqrt(2)    for k=1..7
   */
  i = 0;
  for (row = 0; row < DCTSIZE; row++) 
    for (col = 0; col < DCTSIZE; col++) 
      {
        table[i]=table[i]*aanscalefactor[row] * aanscalefactor[col];
        i++;
      }
  
  for (j=0;j<myrest->height_in_blocks;j++)
    for (i=0;i<myrest->width_in_blocks;i++)
      for (k=0;k<DCTSIZE2;k++)
      {
        adr = (j*myrest->width_in_blocks + i)*64 + k;
        myrest->DCTmin[adr]=((RCOEF) blockarray[j][i][k]-0.5)*table[k];
        myrest->DCTmax[adr]=((RCOEF) blockarray[j][i][k]+0.5)*table[k];
        myrest->DCTimage[adr]=(myrest->DCTmin[adr]+myrest->DCTmax[adr])/2.;
      }
}

/* Function to minimize: E(u) and its gradient dE 
   Weighted-TV adapted to blocking effects.
*/

LOCAL(double)
TV(struct jpeg_restoration *rest, RSAMPBUFFER ug, RSAMPBUFFER dE)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int x,y,center,left,right,down,up;
  double e,v,vleft,vright,vdown,vup,d1,d2,d3,d4;
  RSAMP *ptr;

  for (x=0,ptr=dE; x<myrest->w*myrest->h; x++) *(ptr++)=0.0;
  e = 0.;

  for(y=1;y<myrest->h-1;y++) 
    for (x=1;x<myrest->w-1;x++)
      {
        center = y*myrest->w+x;
        right=center+1;
        down=center+myrest->w;
        left=center-1;
        up=center-myrest->w;

        vup=ug[up];
        vleft=ug[left];
        vdown=ug[down];
        vright=ug[right];
        v=ug[center];

        d1 = myrest->wTV[x%8]*(vright-v);
        d2 = myrest->wTV[y%8]*(vdown-v);
        d3 = myrest->wTV[(x-1)%8]*(v-vleft);
        d4 = myrest->wTV[(y-1)%8]*(v-vup);

        v = sqrt( (d1*d1+d2*d2+d3*d3+d4*d4) );
        e += v;
        if (v==0.) v=1.;
        dE[center]   += (d3+d4-d1-d2)/v;
        dE[right] += d1/v;
        dE[down] += d2/v;
        dE[up] -= d4/v;
        dE[left] -= d3/v;

      }
  return(e);
}

/* Main algorithm : min TV subject to the constraint in the DCT domain
 */


/* Projection on the constraint - by Froment*/
LOCAL(void)
project(struct jpeg_restoration *rest,
  RSAMPBUFFER out)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;

  int j;
  fdct_image(myrest->DCTimage, out, myrest->w, myrest->h);
    
  for (j=0;j<myrest->h*myrest->w;j++)
    if (myrest->DCTimage[j] > myrest->DCTmax[j])
      myrest->DCTimage[j] = myrest->DCTmax[j];
    else if (myrest->DCTimage[j] < myrest->DCTmin[j])
      myrest->DCTimage[j] = myrest->DCTmin[j];

  fdct_inverse_image(myrest->DCTimage, out, myrest->w, myrest->h);

}

LOCAL(void)
min_TV(j_decompress_ptr cinfo, struct jpeg_restoration *rest, int niterations,
    double step, boolean cnt_step, RSAMPBUFFER out)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j;

  double tv;
  double t;
  RSAMPBUFFER dE,p_dE;
  RSAMPBUFFER p_out;

  dE = (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo, myrest->h * myrest->w);

  if (dE == NULL)
    ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 0);

  for (i=0;i<niterations;i++)
    /* Main loop */
    {
      /* compute TV and its gradient */
      tv = TV(rest,out,dE)/(double)(myrest->h*myrest->w);

      /* stepsize */
      if (cnt_step)
        t=step; /* constant step size, to speed-up the algorithm */
      else
        t = step/((double) i+1.0); /* step size that ensures the convergence */

      /*
      fprintf(stderr, "Iteration %i/%i. TV: %g. Step: %g\n", i, niterations,
          tv, t);
          */

      project(rest, out);

      /* Gradient descent */
      for (p_out=out, p_dE=dE, j=0; j<myrest->w*myrest->h; j++,p_out++,p_dE++)
        *p_out -= (float) (t * *p_dE);



    } /* End of main loop */

}
